#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np

from redcp import * # import data types

DELTA_DISTANCE = 0.12
C_TURNING_SPEED = 5
C_AVOID_SPEED = 10

class Controller:
    def __init__(self):
        pass

    def goto(self, goal: Pose, odometry: Odometry, collision: bool) -> Twist:
        """Method to steer the robot towards the goal position given its current 
           odometry and collision status
        Args:
            goal: Pose of the robot goal
            odometry: Perceived odometry of the robot
            collision: bool of the robot collision status
        Returns:
            cmd: Twist steering command
        """
        # zero velocity steering command
        if odometry is None or goal is None: return Twist()
        if collision: return None

        cmd_msg = Twist()
        
        ## get position diff for distance calculation
        dx = goal.position.x - odometry.pose.position.x
        dy = goal.position.y - odometry.pose.position.y
        
        gphi = math.atan2(dy, dx) ## phi from robot to goal
        rphi = odometry.pose.orientation.to_Euler()[0] ## current robot orientation
        dphi = ((gphi-rphi + math.pi) % (2*math.pi)) - math.pi ## phi difference eddited with Milos conversion

        dist = odometry.pose.dist(goal)
        
        ## set message values for robot to move
        cmd_msg.linear.x = dist
        cmd_msg.angular.z = dphi * C_TURNING_SPEED

        
        if dist < DELTA_DISTANCE: ## check if distance under threshold
            return None

        return cmd_msg            

    def goto_reactive(self, goal: Pose, odometry: Odometry, collision: bool, laser_scan: LaserScan) -> Twist:
        """Method to steer the robot towards the goal position while avoiding 
           contact with the obstacles given its current odometry, collision 
           status and laser scan data
        Args:
            goal: Pose of the robot goal
            odometry: Perceived odometry of the robot
            collision: bool of the robot collision status
            laser_scan: LaserScan data perceived by the robot
        Returns:
            cmd: Twist steering command
        """

        if odometry is None or goal is None or laser_scan is None: return Twist()
        if collision: return None

        #zero velocity steering command
        cmd_msg = self.goto(goal, odometry, collision)
        if cmd_msg is None: return None

        n = len(laser_scan.distances)

        ## left points: min_angle + i*increment <= 0
        left_points = [ laser_scan.distances[i] for i in range(n) 
                            if  laser_scan.distances[i] >= laser_scan.range_min 
                            and laser_scan.distances[i] <= laser_scan.range_max
                            and laser_scan.angle_min + i * laser_scan.angle_increment <= 0 ]

        ## right points: min_angle + i*increment > 0
        right_points = [ laser_scan.distances[i] for i in range(n) 
                            if  laser_scan.distances[i] >= laser_scan.range_min 
                            and laser_scan.distances[i] <= laser_scan.range_max
                            and laser_scan.angle_min + i * laser_scan.angle_increment > 0 ]

        closest_scan_left  = 1 if len(left_points) == 0 else np.min(left_points)
        closest_scan_right = 1 if len(right_points) == 0 else np.min(right_points)

        # rotation z axis ~~ angular.z
        repulsive_force = (1 / closest_scan_left) - (1 / closest_scan_right)

        cmd_msg.angular.z += repulsive_force * C_AVOID_SPEED

        return cmd_msg
