#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import math
import time
import copy
import numpy as np

import matplotlib.pyplot as plt

import scipy.ndimage as ndimg
from sklearn.cluster import KMeans

import collections
import heapq

from queue import PriorityQueue

from redcp import * # import data types

from bresenham import bresenham_line

class Planner:

    CELL_FREE = 0.4 ## < value of free cell
    CELL_MID = 0.5
    CELL_OCCU = 0.6 ## > value of occupied cell

    def __init__(self):
        pass

    def prep_mask(self, robot_size, resolution):
        mask = np.mgrid[
               -robot_size:robot_size + 1,
               -robot_size:robot_size + 1,
               ].reshape(2, -1)
        mask = mask[:, np.sqrt((mask[0]) ** 2 + (mask[1]) ** 2) < robot_size]
        return mask

    def grow_obstacles(self, grid_map: OccupancyGrid, robot_size) -> OccupancyGrid:
        """ Method to grow the obstacles to take into account the robot embodiment
        Args:
            grid_map: OccupancyGrid - gridmap for obstacle growing
            robot_size: float - size of the robot
        Returns:
            grid_map_grow: OccupancyGrid - gridmap with considered robot body embodiment
        """
        grid_map_grow = grid_map.clone()

        if grid_map is None or robot_size is None:
            return grid_map_grow

        robot_size = int(robot_size / grid_map_grow.resolution)
        mask = self.prep_mask(robot_size, grid_map_grow.resolution)

        grid = grid_map_grow.data.reshape((grid_map_grow.height, grid_map_grow.width))
        grid[grid >= self.CELL_MID] = 1 # force obstacles
        for point in np.argwhere(grid > self.CELL_OCCU):  # find obstacles then grow around
            y, x = point
            if robot_size <= x < grid_map_grow.width - robot_size and robot_size <= y < grid_map_grow.height - robot_size:
                points = point[:, np.newaxis] + mask
                grid[(*points,)] = 1

        grid[:robot_size + 1] = 1
        grid[:, :robot_size + 1] = 1
        grid[grid_map_grow.height - robot_size:] = 1
        grid[:, grid_map_grow.width - robot_size:] = 1
        grid[grid < self.CELL_OCCU] = 0 # force free space
        return grid_map_grow

    def successor(self, grid_map, state):
        x, y = state
        children = [(x + 1, y + 1), (x + 1, y), (x + 1, y - 1),
                    (x, y + 1), (x, y - 1),
                    (x - 1, y + 1), (x - 1, y), (x - 1, y - 1)]
        children = [c for c in children if 0 <= c[0] < grid_map.width and 0 <= c[1] < grid_map.height]
        grid = grid_map.data.reshape((grid_map.height, grid_map.width))
        children = [c for c in children if grid[c[1], c[0]] < self.CELL_MID]
        return children

    def euclidean(self, a, b):
        return np.sqrt((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2)

    def heuristic(self, a, b):
        return self.euclidean(a, b)

    def plan_path(self, grid_map: OccupancyGrid, start: Pose, goal: Pose) -> Path:
        """ Method to plan the path from start to the goal pose on the grid
        Args:
            grid_map: OccupancyGrid - grid map for planning
            start: Pose - robot start pose
            goal: Pose - robot goal pose
        Returns:
            path: Path - path between the start and goal Pose on the map
        """

        path = Path()

        costs = {}
        predecessor = {}

        q = PriorityQueue()
        s = grid_map.world_to_map(start)
        g = grid_map.world_to_map(goal)
        q.put((0, s))
        costs[s] = 0
        predecessor[s] = None

        while not q.empty():
            priority, cur = q.get()
            if cur == g:
                break

            for child in self.successor(grid_map, cur):
                child_cost = costs[cur] + self.euclidean(cur, child)
                if child not in costs or costs[child] > child_cost:
                    costs[child] = child_cost
                    predecessor[child] = cur
                    q.put((child_cost + self.heuristic(child, g), child))

        if cur != g:
            # print("Goal not found")
            return None
        s = g
        while s is not None:
            path.poses.append(grid_map.map_to_world(s))
            s = predecessor[s]

        path.poses = list(reversed(path.poses))
        path.poses[0] = start
        path.poses[-1] = goal
        return path

    def simplify_path(self, grid_map: OccupancyGrid, path: Path) -> Path: 
        """ Method to simplify the found path on the grid
        Args:
            grid_map: OccupancyGrid - gridmap for obstacle growing
            path: Path - path to be simplified
        Returns:
            path_simple: Path - simplified path
        """
        if grid_map is None or path is None:
            return None

        grid = grid_map.data.reshape((grid_map.height, grid_map.width))

        # get poses as tuples in map coords
        poses = [grid_map.world_to_map(p) for p in path.poses]

        path_simplified = Path()

        # add the start pose
        path_simplified.poses.append(poses[0])

        i = 0
        while i < len(poses)-1:
            last_i = i
            last_p = poses[i]
            i += 1
            for j in range(i+2,len(poses)):
                p = poses[j]
                pts = np.array(tuple(bresenham_line(last_p, p)), dtype=np.int).T
                if np.all(grid[pts[1,:], pts[0,:]] <= self.CELL_OCCU):  # no collision
                    i = j
                else:  # collision
                    break
            path_simplified.poses.append(poses[i])        

        path_simplified.poses = [grid_map.map_to_world(p) for p in path_simplified.poses]
        return path_simplified

