#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import math
import copy
import numpy as np

from redcp import * # import data types

from bresenham import bresenham_line

P_FREE = 0.1
P_OCC = 0.9
P_MIN = 0.01

class Mapper:

    def __init__(self):
        pass

    def update_free(self, p_mi):
        """method to calculate the Bayesian update of the free cell with the current occupancy probability value P_mi
        Args:
            p_mi: float64 - current probability of the cell being occupied
        Returns:
            p_mi: float64 - updated probability of the cell being occupied
        """
        p = p_mi * P_FREE
        p_mi = p / (2 * p - P_FREE - p_mi + 1)
        if p_mi < P_MIN: p_min = P_MIN
        if p_mi > (1 - P_MIN): p_mi = (1 - P_MIN)
        return p_mi

    def update_occupied(self, p_mi):
        """method to calculate the Bayesian update of the occupied cell with the current occupancy probability value P_mi
        Args:
            p_mi: float64 - current probability of the cell being occupied
        Returns:
            p_mi: float64 - updated probability of the cell being occupied
        """
        p = p_mi * P_OCC
        p_mi = p / (2 * p - P_OCC - p_mi + 1)
        if p_mi < P_MIN: p_min = P_MIN
        if p_mi > (1 - P_MIN): p_mi = (1 - P_MIN)
        return p_mi

    def fuse_laser_scan(self, grid_map: OccupancyGrid, laser_scan: LaserScan, odometry: Odometry) -> OccupancyGrid:
        """ Method to fuse the laser scan data sampled by the robot with a given 
            odometry into the probabilistic occupancy grid map
        Args:
            grid_map: OccupancyGrid - gridmap to fuse te laser scan to
            laser_scan: LaserScan - laser scan perceived by the robot
            odometry: Odometry - perceived odometry of the robot
        Returns:
            grid_map_update: OccupancyGrid - gridmap updated with the laser scan data
        """
 
        if grid_map is None: ## check if input params are valid
            return None
            
        grid_map_update = copy.deepcopy(grid_map)
        # if laser_scan is None or odometry is None or grid_map.data is None:
        if laser_scan is None or odometry is None:
            return grid_map_update

        r_ori = odometry.pose.orientation.to_Euler()[0] ## robot heading
        R = np.array([[math.cos(r_ori), -math.sin(r_ori)], ## rotation matrix
                      [math.sin(r_ori),  math.cos(r_ori)]])
        T = np.array([odometry.pose.position.x, odometry.pose.position.y]) ## translation vector

        ## get coordinates of origin as index for matrix
        origin_mx = round((odometry.pose.position.x - grid_map.origin.position.x) / grid_map.resolution)
        origin_my = round((odometry.pose.position.y - grid_map.origin.position.y) / grid_map.resolution)
        start = (origin_mx, origin_my)

        ## gridmap transformed to matrix instead of vector
        data = grid_map_update.data.reshape(grid_map_update.height, grid_map_update.width)

        for i in range(len(laser_scan.distances)): 
            angle = laser_scan.angle_min + laser_scan.angle_increment * i ## compute x and y of the scan
            p_i = laser_scan.distances[i] * np.array([math.cos(angle), math.sin(angle)]) ## get scan coordinates with respect to robots view
            o_i = np.matmul(R, p_i) + T ## transform to odometry view, new shift with origin
            o_i -= np.array([grid_map.origin.position.x, grid_map.origin.position.y])

            mx = round(o_i[0] / grid_map.resolution) ## calculate indexes in the data matrix
            my = round(o_i[1] / grid_map.resolution)

            goal = (mx, my)
            pts = bresenham_line(start, goal)
            
            for pt in pts: ## Free points
                if pt[0] == goal[0] and pt[1] == goal[1]: 
                    continue  ## should not happen anyways
                if pt[0] >= grid_map.width or pt[1] >= grid_map.height: 
                    continue  ## trim points
                if pt[0] < 0 or pt[1] < 0: 
                    continue

                data[pt[1], pt[0]] = self.update_free(data[pt[1], pt[0]])

            ## Occupied point (only goal)
            if goal[0] >= grid_map.width or goal[1] >= grid_map.height: 
                continue
            if goal[0] < 0 or goal[1] < 0: 
                continue
            data[goal[1], goal[0]] = self.update_occupied(data[goal[1], goal[0]])

        ## transform data back to vector
        grid_map_update.data = data.flatten()
        return grid_map_update
